"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.brain = exports.bf = exports.readUTF32Char = void 0;
var utf32char_1 = require("utf32char");
var Token_1 = require("./Token");
var readline = require('readline');
var readlineSync = require('readline-sync');
function readUTF32Char() {
    var input = "";
    while (true) {
        try {
            input = readlineSync.question("> ");
            return utf32char_1.UTF32Char.fromString(input);
        }
        catch (error) {
            console.log(error.message);
            console.log("Not a valid UTF-32 character: \"" + input + "\"");
            console.log("  Please try again, or press CTRL-C to quit");
        }
    }
}
exports.readUTF32Char = readUTF32Char;
function bf(program, classic, memory, input) {
    if (memory === void 0) { memory = utf32char_1.UInt32.fromNumber(1000); }
    if (input === void 0) { input = readUTF32Char; }
    if (memory.lt(1))
        throw new Error("must provide at least 1 cell of memory");
    var state = new Array(memory.toNumber()).fill(utf32char_1.UInt32.fromNumber(0));
    var memoryPointer = new utf32char_1.UInt32(0);
    var programPointer = new utf32char_1.UInt32(0);
    var output = "";
    function movePointer(op) {
        if (op === '<') {
            if (classic) {
                if (memoryPointer.lt(1))
                    memoryPointer = memory.minus(1);
                else
                    memoryPointer = memoryPointer.minus(1);
            }
            else {
                try {
                    memoryPointer = memoryPointer.minus(1);
                }
                catch (error) {
                    throw new Error("cannot decrement memory pointer below minimum index 0... try setting `classic` to `true`");
                }
            }
        }
        else if (op === '>') {
            if (classic) {
                if (memoryPointer.ge(memory.minus(1)))
                    memoryPointer = utf32char_1.UInt32.fromNumber(0);
                else
                    memoryPointer = memoryPointer.plus(1);
            }
            else {
                if (memoryPointer.ge(memory.minus(1)))
                    throw new Error("cannot increment memory pointer above maximum index " + memory.minus(1).toString() + "... try setting `classic` to `true`");
                else
                    memoryPointer = memoryPointer.plus(1);
            }
        }
        else
            throw new Error("movePointer received invalid op: " + op);
    }
    function changeCell(op) {
        if (!Token_1.isToken(op))
            throw new Error("changeCell received invalid op: " + op);
        var index = memoryPointer.toNumber();
        var value = state[index];
        if (op === '+') {
            if (classic) {
                if (value.ge(255))
                    state[index] = utf32char_1.UInt32.fromNumber(0);
                else
                    state[index] = value.plus(1);
            }
            else {
                if (value.ge(utf32char_1.UInt32.MAX_VALUE))
                    throw new Error("cell " + index + " is already at maximum allowable value, 2^32 - 1... try setting `classic` to `true`");
                else
                    state[index] = value.plus(1);
            }
        }
        else if (op === '-') {
            if (classic) {
                if (value.lt(1))
                    state[index] = utf32char_1.UInt32.fromNumber(255);
                else
                    state[index] = value.minus(1);
            }
            else {
                if (value.le(utf32char_1.UInt32.MIN_VALUE))
                    throw new Error("cell " + index + " is already at minimum allowable value, 0... try setting `classic` to `true`");
                else
                    state[index] = value.minus(1);
            }
        }
        else {
            throw new Error("changeCell received invalid op: " + op);
        }
    }
    function mayFail(op) {
        try {
            op();
        }
        catch (err) {
            var padWith = "              ";
            var padding = padWith.length;
            var ptr = programPointer.toNumber();
            var minChIncl = Math.max(0, ptr - padding);
            var maxChExcl = Math.min(ptr + padding, program.length);
            var padLeft = padWith.slice(ptr - padding);
            var errString = "\n" +
                padLeft + program.slice(minChIncl, maxChExcl) + '\n' +
                ("Error at      ^      char index " + ptr) + '\n' +
                ("  message: " + err.message) + '\n' +
                ("   output: " + output) + '\n';
            throw new Error(errString);
        }
    }
    function jump(op) {
        if (!Token_1.isToken(op))
            throw new Error("moveToMatch received invalid op: " + op);
        var delta = 1;
        var index = 1;
        if (op === '[') {
            var chars = program.slice(programPointer.toNumber() + 1).split("");
            for (var _i = 0, chars_1 = chars; _i < chars_1.length; _i++) {
                var char = chars_1[_i];
                if (char === ']')
                    delta -= 1;
                else if (char === '[')
                    delta += 1;
                if (delta > 0)
                    index += 1;
                else
                    break;
            }
            if (delta > 0)
                throw new Error("missing close-brace ] for this open-brace [");
            else
                programPointer = programPointer.plus(index);
        }
        else if (op === ']') {
            var chars = program.slice(0, programPointer.toNumber()).split("").reverse();
            for (var _a = 0, chars_2 = chars; _a < chars_2.length; _a++) {
                var char = chars_2[_a];
                if (char === '[')
                    delta -= 1;
                else if (char === ']')
                    delta += 1;
                if (delta > 0)
                    index += 1;
                else
                    break;
            }
            if (delta > 0)
                throw new Error("missing open-brace [ for this close-brace ]");
            else
                programPointer = programPointer.minus(index);
        }
        else {
            throw new Error("moveToMatch received invalid op: " + op);
        }
    }
    var _loop_1 = function () {
        var nextChar = program[programPointer.toNumber()];
        if (Token_1.isToken(nextChar)) {
            var index = memoryPointer.toNumber();
            var value = state[index];
            if (nextChar === '[' && value.lt(1)) {
                mayFail(function () { return jump(nextChar); });
            }
            else if (nextChar === ']' && value.gt(0)) {
                mayFail(function () { return jump(nextChar); });
            }
            else if (nextChar === '<' || nextChar === '>') {
                mayFail(function () { return movePointer(nextChar); });
            }
            else if (nextChar === '+' || nextChar === '-') {
                mayFail(function () { return changeCell(nextChar); });
            }
            else if (nextChar === '.') {
                output += utf32char_1.UTF32Char.fromUInt32(value).toString();
            }
            else if (nextChar === ',') {
                state[index] = input().toUInt32();
            }
        }
        programPointer = programPointer.plus(1);
    };
    while (programPointer.toNumber() < program.length) {
        _loop_1();
    }
    return output;
}
exports.bf = bf;
function brain() {
    var mode = false;
    var signoffs = [
        "Totsiens", "Ma'a as-salaama", "Bidāẏa", "Zdravo", "Joigin",
        "Donadagohvi", "Doviđenja", "Sbohem", "Farvel", "Tot ziens",
        "Nägemist", "Näkemiin", "Au Revoir", "Auf Wiedersehen", "Yasou",
        "Aloha", "L'hitraot", "Namaste", "Viszlát", "Vertu sæll",
        "Sampai Jumpa", "Slán", "Arrivederci", "Sayōnara", "안녕",
        "Vale", "Uz redzēšanos", "Atsiprasau", "Zài jiàn", "Ha det bra",
        "Khodaa haafez", "Żegnaj", "Adeus", "Alweda", "La revedere",
        "Прощай", "Dovidenia", "Nasvidenje", "Adios", "Adjö",
        "Poitu varein", "Laa Gòn", "Görüşürüz", "Do pobachennia", "Khuda hafiz",
        "Tạm biệt", "Hwyl fawr", "Hamba kahle", "再见", "وداعا",
        "May the force be with you", "ลาก่อน", "பிரியாவிடை", "Namárië",
        "Qapla'",
        "Live long and prosper"
    ];
    var signoff = "👋 " + signoffs[Math.floor(Math.random() * signoffs.length)];
    var rl = readline.createInterface({
        input: process.stdin,
        output: process.stdout,
        prompt: "\n🧠: "
    });
    console.log("\nEnter single-line BF code below or");
    console.log("  type :paste to paste multiline code");
    console.log("  type :mode to toggle classic / default mode");
    console.log("  type :quit or enter <CTRL>-C or <CTRL>-D to quit");
    var multiLine = "";
    var pasteMode = false;
    var previousLine = "";
    rl.prompt();
    function interpret(program, mode) {
        try {
            console.log(bf(program, mode));
        }
        catch (error) {
            console.log(error.message);
        }
    }
    rl.on('line', function (line) {
        if (pasteMode) {
            if (previousLine === "" && previousLine === line) {
                console.log("~~~~~~~~~~~~~~~ BEGIN OUTPUT ~~~~~~~~~~~~~~\n");
                interpret(multiLine, mode);
                multiLine = "";
                pasteMode = false;
                rl.prompt();
            }
            else {
                previousLine = line;
                multiLine += line;
            }
        }
        else if (line === ":paste" && multiLine === "" && !pasteMode) {
            console.log("\nEntering multiline input mode.");
            console.log("Enter two blank lines in a row to interpret.");
            console.log("~~~~~~~~~~~~~~~ BEGIN INPUT ~~~~~~~~~~~~~~~\n\n");
            pasteMode = true;
        }
        else if (line === ":mode") {
            mode = !mode;
            var modeStr = void 0;
            if (mode)
                modeStr = "classic";
            else
                modeStr = "default";
            console.log("\nChanged interpretation mode to '" + modeStr + "'");
            rl.prompt();
        }
        else if (line === ":quit") {
            rl.prompt();
            console.log(signoff + "\n");
            process.exit(0);
        }
        else {
            if (line.length > 0)
                interpret(line, mode);
            rl.prompt();
        }
    }).on('close', function () {
        console.log(signoff + "\n");
    });
}
exports.brain = brain;
